// -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vim:ts=2:sw=2:tw=80:et

/* 
 * Copyright (c) 2002, Andrew Carra and Robert Hagemann
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the WiGLE.net nor Mimezine nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.mimezine.geo;

import java.util.Iterator;

/**
 * simple class to hold a GPS point.
 * GPSPoint is immutable, you cannot mute it.
 */
public final class GPSPoint {
    
  /** version info */
  public static final String VERSION_ID = 
    "$Id: GPSPoint.java,v 1.6 2003/05/14 14:29:35 uhtu Exp $";

  /** the intersection of the equator and the prime meridian */
  public static final GPSPoint ZERO_POINT = new GPSPoint(0.0f, 0.0f);
  
  /** our latitude */
  private final float lat1;
  
  /** our longitude */
  private final float long1;
  
  /** no default constructor */
  private GPSPoint() {
    this(0f, 0f);
  }
  
  /**
   * Build a GPSPoint for the given latitude and longitude.
   *
   * @param lat the latitude.
   * @param lon the longitude.
   * @throws IllegalArgumentException if lat or lon are out of range.
   */
  public GPSPoint( float lat, float lon ) throws IllegalArgumentException {
    if ( ( ( lat < -90f ) || ( lat > 90f ) )
         ||
         ( ( lon < -180f ) || ( lon > 180f ) ) ) {
      // XXX: in practice this happens a lot, we'll have to look into options
      // for handling these cases
      // throw new IllegalArgumentException("invalid lat/long values");
    }
                                        
    lat1  = lat;
    long1 = lon;
  }

  /**
   * copy constructor. clone() is dangerous.
   *
   * @param point the point to copy.
   * @throws IllegalArgumentException if point is out of range
   */
  public GPSPoint(GPSPoint point) throws IllegalArgumentException {
    this(point.getLat(), point.getLon());
  }


  /**
   * Take the rectangular average of this point and another.
   * 
   * @param point the point to average with.
   * @return a new GPSPoint halfway between this one and point.
   * @throws IllegalArgumentException if point is out of range
   */
  public GPSPoint average( GPSPoint point ) throws IllegalArgumentException {
    return new GPSPoint( (point.lat1 + lat1) / 2,
      (point.long1 + long1) / 2 );
  }

  /**
   * Find the rectangular distance between this point and another.
   * er. in latitude/longitude units. that dosen't even make any sense.
   *
   * @param point the point to calculate distance from.
   * @return the rectangular distance between this one and point.
   */
  public float length( GPSPoint point ) {
    float latdiff  = (point.lat1 - lat1);
    float longdiff = (point.long1 - long1);
    
    // XXX: can this truncation cause problems with our values? 
    return (float) Math.sqrt( (latdiff * latdiff) + (longdiff * longdiff) );
  }



  /*
    from  http://mathforum.org/library/drmath/view/51879.html

    Presuming a spherical Earth with radius R (see below), and that the
    locations of the two points in spherical coordinates (longitude and
    latitude) are lon1,lat1 and lon2,lat2, then the Haversine Formula 
    (from R. W. Sinnott, "Virtues of the Haversine," Sky and Telescope, 
    vol. 68, no. 2, 1984, p. 159): 
    
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = (sin(dlat/2))^2 + cos(lat1) * cos(lat2) * (sin(dlon/2))^2
    c = 2 * atan2(sqrt(a), sqrt(1-a)) 
    d = R * c
    
    R = 6367 km = 3956 mi 
   */

  /** get me a more accurate model of the earth and i will move the world */
  public static final int RADIUS_IN_METERS = 6367 * 1000;

  /**
   * calculate the correct spherical distance between this point and the
   * given point in meters.
   *
   * @param point the point to calculate distance from.
   * @return the distance in meters between this point and the given point.
   */
  public double distance( GPSPoint point ) {
    // XXX: should be checking for NaN in here
    double lat1 = Math.toRadians(this.lat1);
    double lon1 = Math.toRadians(this.long1);
    double lat2 = Math.toRadians(point.lat1);
    double lon2 = Math.toRadians(point.long1);
    double dlon = Math.abs( lon2 - lon1 );
    double dlat = Math.abs( lat2 - lat1 );
    double temp1 = Math.sin(dlat/2);
    temp1 = temp1 * temp1;
    double temp2 = Math.sin(dlon/2);
    temp2 = temp2 * temp2;
    double a = temp1 + Math.cos(lat1) * Math.cos(lat2) * temp2;
    double c = 2 * Math.atan2( Math.sqrt(a), Math.sqrt(1-a) );

    return RADIUS_IN_METERS * c;
  }

  /*
    taken from http://mathforum.org/library/drmath/view/51822.html
  
    lon3 = tan^-1(cos(lat2)*sin(dlon)/
        (cos(lat1)+cos(lat2)*cos(dlon)))

    lat3 = tan^-1((sin(lat1)+sin(lat2))/
                sqrt((cos(lat1)+cos(lat2)*cos(dlon))^2+
                     (cos(lat2)*sin(dlon))^2))
   */
  /**
   * Finds the correct spherical coordinate midway between this point
   * and the given point.
   *
   * @param point the point to find the midpoint between 
   * @return a new GPSPoint halfway between this one and the given point.
   * @throws IllegalArgumentException if point is out of range
   */
  public GPSPoint midpoint( GPSPoint point ) throws IllegalArgumentException {
    // XXX: should be checking for NaN in here
    double lat1 = Math.toRadians(this.lat1);
    double lon1 = Math.toRadians(this.long1);
    double lat2 = Math.toRadians(point.lat1);
    double lon2 = Math.toRadians(point.long1);
    double dlon = Math.abs( lon2 - lon1 );
    double dlat = Math.abs( lat2 - lat1 );
    
    double lon3 = Math.atan( ( Math.cos(lat2) * Math.sin(dlon) ) /
      ( Math.cos(lat1) + Math.cos(lat1) * Math.cos(dlon) ) );
    double temp1 = Math.cos(lat1) + Math.cos(lat2) * Math.cos(dlon);
    temp1 = temp1 * temp1;
    double temp2 = Math.cos(lat2) * Math.sin(dlon);
    temp2 = temp2 * temp2;

    double lat3 = Math.atan( ( Math.sin(lat1) + Math.sin(lat2) ) /
                             Math.sqrt( temp1 + temp2 ) );
    return new GPSPoint( (float)lat3, (float)lon3 );
  }


  /**
   * @return the latitude
   */
  public float getLat() {
    return lat1;
  }

  /**
   * @return the longitude
   */
  public float getLon() {
    return long1;
  }

  /**
   * @return is this point west of the prime meridian?
   */
  public boolean isWest() {
    return (0f > long1);
  }

  /**
   * @return is this point east of the prime meridian?
   */
  public boolean isEast() {
    return (0f < long1);
  }
  
  /**
   * @return is this point on the prime meridian?
   */
  public boolean isPrimeMeridan() {
    return (0f == long1);
  }

  /**
   * @return is this point north of the equator?
   */  
  public boolean isNorth() {
    return (0f < lat1);
  }
  
  /**
   * @return is this point south of the equator?
   */
  public boolean isSouth() {
    return (0f > lat1);
  }
  
  /**
   * @return is this point on the equator?
   */
  public boolean isEquator() {
    return (0f == lat1);
  }


  public String toString() {
    return "(" + lat1 + "," + long1 + ")";
  }


  /**
   * @param points a Collection of GPSPoints
   * @param stddev the number of Standard deviations allowed to be returned
   * @return a Collection of GPSPoints within stddev standard deviations from points
   */
  public static final java.util.Collection filterByStdDev( java.util.Collection points, int stddev ) {
    // calculate mean for lat,lon
    java.util.List l = new java.util.LinkedList();

    int count = 0;
   
    // XXX: overflow?
    double lattotal = 0f;
    double lontotal = 0f;
    for ( Iterator i = points.iterator(); i.hasNext(); count++ ) {
      GPSPoint p = (GPSPoint) i.next();
      lattotal += p.getLat();
      lontotal += p.getLon();
    }
    if ( Double.isInfinite(lattotal) || Double.isInfinite( lontotal )  ) {
       throw new ArithmeticException("double overflow in mean");
    }
    if ( count == 0 ) {
       return l;
    }

    float latmean = (float) (lattotal / count);
    float lonmean = (float) (lontotal / count);

    double latsum = 0f;
    double lonsum = 0f; // all by yer lonsum

    // calculate std deviation for lat, lon

    for ( Iterator i = points.iterator(); i.hasNext(); ) {
      GPSPoint p = (GPSPoint) i.next();
      latsum += Math.pow( p.getLat() - latmean, 2.0 );
      lonsum += Math.pow( p.getLon() - lonmean, 2.0 );
    }

    double latstddev = Math.sqrt( latsum / count ); // use count-1 for unbiased sample variance
    double lonstddev = Math.sqrt( lonsum / count ); // use count-1 for unbiased sample variance

    if ( Double.isInfinite( latstddev ) || Double.isInfinite( lonstddev )  ) {
      throw new ArithmeticException("double overflow in stddev");
    }

    //    System.err.println("lat mean: "+latmean+" lon mean: "+lonmean+
    //                       " lat stdd: "+latstddev+" lon stdd: "+lonstddev );

    float nlatstddev = (float) (latstddev * stddev);
    float nlonstddev = (float) (lonstddev * stddev);

     // XXX: this will always be correct, excepting the dateline edge case issue
    GPSQuad bounds = new GPSQuad( latmean + nlatstddev, lonmean - nlonstddev, 
                                  latmean - nlatstddev, lonmean + nlonstddev );

    // filter out members of points outside stddev * (that range)
    for ( Iterator i = points.iterator(); i.hasNext(); ) {
      GPSPoint p = (GPSPoint) i.next();
      if ( bounds.contains( p ) ) {
        l.add(p);
      }
    }

    return l;
  }
  

  // per item 7 from effective java
  // obey the general contract when overriding equals
  public boolean equals( Object obj ) {
    if ( this == obj ) {
      return true;
    }
    
    if ( !(obj instanceof GPSPoint) ) {
      return false;
    }
    
    GPSPoint point = (GPSPoint) obj;
    
    // check members using Float.floatToIntBits() per 
    // http://java.sun.com/j2se/1.3/docs/api/java/lang/Float.html#equals(java.lang.Object)
    return 
      ( Float.floatToIntBits(lat1) == Float.floatToIntBits(point.lat1)   ) &&
      ( Float.floatToIntBits(long1) == Float.floatToIntBits(point.long1) );
  }
  
  // per item 8 from effective java
  // always override hashcode when you override equals
  public int hashCode() {
    int result = 17;
    result = 37 * result + Float.floatToIntBits(lat1);
    result = 37 * result + Float.floatToIntBits(long1);
    return result;
  }
}
