// -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vim:ts=2:sw=2:tw=80:et

/* 
 * Copyright (c) 2002-2003, Andrew Carra and Robert Hagemann
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the WiGLE.net nor Mimezine nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.mimezine.geo;

/** 
 * simple class to hold 2 sets of GPS points that bound a rectangular region.
 * GPSQuad is immutable.
 */
public final class GPSQuad {
  public static final String VERSION_ID =
    "$Id: GPSQuad.java,v 1.5 2003/02/22 19:58:56 uhtu Exp $";

  /** bounding box for the United States */
  public static final GPSQuad BOUNDS_US = new GPSQuad( 24.39630890f,
    -66.88507843f, 49.38435745f, -124.70354462f );

  /** first corner */
  private final GPSPoint point1;
  /** second corner */
  private final GPSPoint point2;

  /** do not instantiate */
  private GPSQuad () {
    this( GPSPoint.ZERO_POINT, GPSPoint.ZERO_POINT );
  }

  /** 
   * build a new GPSQuad.
   *
   * @param lat1 first corner latitude
   * @param long1 first corner longitude
   * @param lat2 second corner latitude
   * @param long2 second corner longitude
   */
  public GPSQuad ( float lat1, float long1, float lat2, float long2 ) {
    this( new GPSPoint( lat1, long1 ), new GPSPoint( lat2, long2 ) );
  }

  /** 
   * build a new GPSQuad.
   *
   * @param point1 first corner
   * @param point2 second corner
   */ 
  public GPSQuad ( GPSPoint point1, GPSPoint point2 ) {
    this.point1 = point1;
    this.point2 = point2;
  }

  /**
   * copy constructor.
   * @param g the point to copy
   */
  public GPSQuad(GPSQuad g) {
    this( g.point1, g.point2 );
  }

  /**
   * @return first corner latitude
   */
  public float lat1() {
    return point1.getLat();
  }

  /**
   * @return first corner longitude
   */
  public float long1() {
    return point1.getLon();
  }

  /**
   * @return second corner latitude
   */
  public float lat2() {
    return point2.getLat();
  }

  /**
   * @return second corner longitude
   */
  public float long2() {
    return point2.getLon();
  }

  /**
   * @return get first point
   */
  public GPSPoint getPoint1() {
    return point1;
  }

  /**
   * @return get second point
   */
  public GPSPoint getPoint2() {
    return point2;
  }

  /**
   * @return does this GPSQuad contain the point p?
   */
  public boolean contains(GPSPoint p) {
    
    // we have a bounding box (point1, point2)
    // how to best determine inclusion of point p?

    // see if p lon is inside of the two points
    // we don't guarantee ordering, so have to check both ways
    // 1) is the point between the longitude of the two corners (either way)
    // 2) is the point between the latitude of the two corners (either way)
    // XXX: there has got to be a better way of stating this...
    boolean b = 
      ( ( ( p.getLon() >= point1.getLon() ) && 
          ( p.getLon() <= point2.getLon() ) )
        ||
        ( ( p.getLon() <= point1.getLon() ) &&
          ( p.getLon() >= point2.getLon() ) ) )
      &&
      ( ( ( p.getLat() >= point1.getLat() ) && 
          ( p.getLat() <= point2.getLat() ) )
        ||
        ( ( p.getLat() <= point1.getLat() ) &&
          ( p.getLat() >= point2.getLat() ) ) );    

    //    System.err.println(""+this+" contains point "+p+" :"+b);

    return b;
  }


  public String toString() {
    return "(" + lat1() + "," + long1() + "),(" + lat2() + "," + long2() + ")";
  }

  // per item 7 from effective java
  // obey the general contract when overriding equals
  public boolean equals( Object obj ) {
    if ( this == obj ) {
      return true;
    }
    
    if ( !(obj instanceof GPSQuad) ) {
      return false;
    }
    
    GPSQuad q = (GPSQuad) obj;

    // delegate
    return ( ( point1 == null ? q.point1 == null : point1.equals(q.point1) ) &&
             ( point2 == null ? q.point2 == null : point2.equals(q.point2) ) );
  }
  
  // per item 8 from effective java
  // always override hashcode when you override equals
  public int hashCode() {
    int result = 17;
    if ( point1 != null ) {
      result = 37 * result + point1.hashCode();
    }
    if ( point2 != null ) {
      result = 37 * result + point2.hashCode();
    }
    return result;
  }
 
}
